@@ -22,10 +22,17 @@ import ROOT
 import sys
 # import root_numpy# 2019 is old with newer numpy
 
+import psutil
 
 #================ OPERATIONS ON NAMES, INFO ON COLS ROWS
 #------------------------------------------------------------------------
 
+def freemem():
+    a = psutil.virtual_memory()
+    return f"D... AVAIL MEM: {a.available/1024/1024/1024:.2f} GB  ...  used: {a.percent} %"
+
+#------------------------------------------------------------------------
+
 def filename_decomp(filename):
     """
     The ONLY procedure to get the start time from filename
@@ -151,7 +158,7 @@ def column_to_histo( dfcol , binmax = 32*1024, savename = "", hname = "h1", writ
     - saves txt
     - updates allhist.root too
     """
-    print("D... column to_numpy")
+    print("D... creating histo .... column to_numpy")
     narr = dfcol.to_numpy()
     print("D... len=", len(narr), narr.dtype, ".. now to histo")
     his,bins = np.histogram(narr,bins=binmax,range=(0.0,binmax) )
@@ -199,6 +206,30 @@ def save_to_tree(df2, filename):
 
 #========================== READ; ANALYSIS AND DETECTION ===================
 
+def pd_link_next(df):
+    print( freemem() )
+ #
+    print(f"D... broadening table - dtus")
+    df['dtus'] = (df.time.shift(-1)-df.time)*1000*1000
+    df.fillna(0)
+    df['dtus'] = df['dtus'].astype('float32') #
+    print(df)
+    print( freemem() )
+
+
+    print(f"D... broadening table - next_E")
+    df['next_E'] = (df.E.shift(-1))
+    df.fillna(0, inplace=True)
+    df['next_E'] = df['next_E'].astype('int32') #
+    print( freemem() )
+
+    # print(df) # seems to stay the same
+    print(df)
+    print(df.dtypes)
+    print( freemem() )
+
+    return df
+
 #------------------------------------------------------------------------
 def pd_detect_zeroes(df, channel,  TIMEWIN_US = 25  ):
     """
@@ -210,16 +241,29 @@ def pd_detect_zeroes(df, channel,  TIMEWIN_US = 25  ):
       3rd phase - create
     """
 
-    print("D... MEMORY", df.memory_usage())
+    print( freemem() )
     # CHANNEL (i hope it is sorted)
     df = df.loc[ df["ch"] == channel ]
+    if len(df) ==0 :
+        return 1,0
     #
     print(f"D... broadening table - dt")
-    df['dtus'] = (df.time-df.time.shift())*1000*1000
-    print("D... MEMORY", df.memory_usage())
+    df['dtus'] = (df.time.shift(-1)-df.time)*1000*1000
+    df.fillna(0)
+    df['dtus'] = df['dtus'].astype('float32') #
+    print(df)
+    print( freemem() )
+
+
     print(f"D... broadening table - dt")
-    df['next_E'] = (df.E.shift())
-    print("D... MEMORY", df.memory_usage())
+    df['next_E'] = (df.E.shift(-1))
+    df.fillna(0, inplace=True)
+    df['next_E'] = df['next_E'].astype('int32') #
+    print( freemem() )
+
+    # print(df) # seems to stay the same
+    print(df)
+    print(df.dtypes)
 
     #df['dtus'] = (df.t-df.t.shift())*1000*1000
     #dfm = pd.concat([df, df.shift(-1).add_prefix('next_')], axis=1)
@@ -244,24 +288,25 @@ def pd_detect_zeroes(df, channel,  TIMEWIN_US = 25  ):
     # #--- df is prepared
 
 
-
+    print("D... searching for correlations")
     df_dz = df.loc[ ((df.E==0) & (df.next_E==0)) & (df.dtus<TIMEWIN_US) ]
     dzeroes = len(df_dz)
+    print(freemem() )
 
     df_sz = df.loc[ ((df.E!=0) & (df.next_E==0)) & (df.dtus<TIMEWIN_US) ]
     szeroes = len(df_sz)
+    print(freemem() )
 
     df_iz = df.loc[ ((df.E==0) & (df.next_E!=0)) & (df.dtus<TIMEWIN_US) ]
     izeroes = len(df_iz)
+    print(freemem() )
 
 
     print(f"D... ---------------------- START of zero detection ch={channel}")
     df_z = df[ (df.E==0) & (df.ch==channel)]
     zeroes = len(df_z)
+    print(freemem() )
 
-    # print(df) # seems to stay the same
-    print(df)
-    print("D... MEMORY", df.memory_usage())
 
     print(f"D... total  zeroes (chan={channel}) = {zeroes} which is  {zeroes/len(df)*100:5.2f}%")
     print(f"D... double zeroes (chan={channel}) = {dzeroes} which is  {dzeroes/len(df)*100:5.2f}%")
@@ -272,6 +317,7 @@ def pd_detect_zeroes(df, channel,  TIMEWIN_US = 25  ):
     print(df_iz)
     print("D... ______________________ end of zero detection")
 
+
     # returns length for now
     return zeroes,dzeroes
 
@@ -281,8 +327,28 @@ def pd_read_table(filename, sort = True):
     """
     READ THE ASC TABLE:   two possibilities:  4 columns OR 5 columns
     """
-    df = pd.read_table( filename, names=['time','E','x','ch','y'], sep = "\s+", comment="#")
-    df['time'] = df['time']/1e+8
+    df = pd.read_table( filename,
+                        names=['time','E','x','ch','y'],
+                        sep = "\s+",
+                        comment="#")
+
+    print(freemem() )
+    # drop if 4 column
+    df = df.dropna(axis='columns', how='all')
+
+
+    df['time'] = df['time']/1e+8 # this is correct for 10ns step ADC
+
+    #print(df.dtypes)
+    df['time'] = df['time'].astype('float64')
+    df['E']    = df['E'].astype('int32')
+    df['x']    = df['x'].astype('int32')
+    df['ch']   = df['ch'].astype('int32')
+    if 'y' in df:
+        df['y'] = df['y'].astype('int32')
+    print("D...    finale types ************************************")
+    print(df.dtypes)
+    # print(freemem())
 
     #    df.to_hdf('run'+str(number)+'.h5',
     #              "matrix",
@@ -290,6 +356,7 @@ def pd_read_table(filename, sort = True):
     #              data_columns=True,
     #              mode='w')
 
+    # print(freemem() )
     if sort:
         print("D... sorting (after read_table):")
         #print(df)
@@ -297,12 +364,14 @@ def pd_read_table(filename, sort = True):
         #print(df)
         df.reset_index(inplace=True, drop=True)
         #print(df)
+        #print( freemem() )
 
     return df
 
 
 
 
+
 #====================================== MAIN ==================================
 #------------------------------------------------------------------------
 
